//
//  ToDoListTableViewController.swift
//  ToDoList
//
//  Created by Stephen Freund on 11/18/18.
//  Copyright © 2018 Stephen Freund. All rights reserved.
//

import UIKit

/**
 Client App for our ToDoList REST service.
 */
class ToDoListTableViewController: UITableViewController {
  
  /// The port/host of the server we're connecting to.
  private let port = 9080
  private let host = "localhost"
  
  // MARK: - Model
  
  /// The items in our to do list.
  private var items : [ToDoItem] = [] {
    didSet {
      tableView.reloadData()
    }
  }

  // MARK: - ViewController Lifecycle

  /**
   Set up the controller with the initial list of todo items.
   
   **Modifies**: Self
   
   **Effects**: Loads items from the server.
   */
  override func viewDidLoad() {
    super.viewDidLoad()
    self.navigationItem.rightBarButtonItem = self.editButtonItem
    self.loadItems()
  }
  
  // MARK: - Methods communicating with server
  
  /**
   Get the todo list from the server, and update our model.
   
   **Modifies**: Self
   
   **Effects*: Changes items to reflect what is on the server, and also
   changes the title of the controller to reflect how many items
   we have.
   
   */
  private func loadItems() {
    if let url = URL(string: "http://\(self.host):\(self.port)/todos") {
      // since the request may take a while, run in global queue
      DispatchQueue.global().async { [weak self] in
        if let data = try? Data(contentsOf: url) {
          // switch back to main queue to update UI
          DispatchQueue.main.async {
            if let items = try? JSONDecoder().decode([ToDoItem].self, from: data) {
              self?.items = items
              self?.title = "To Do: \(items.count) Items"
            } else {
              self?.items = []
              self?.title = "To Do: Can't Get Items"
            }
            self?.tableView.refreshControl?.endRefreshing()
          }
        }
      }
    }
  }
  
  /**
   Add a new non-completed task with the current time stamp.
  
   **Modifies**: Self
   
   **Effects**: Adds the task to our list on the server, and updates our table.
   
   -Parameter task: The todo item to add.
   */
  private func add(task: String) {
    // Format the date nicely
    let formatter = DateFormatter()
    formatter.timeStyle = .short
    formatter.dateStyle = .short
    let date = formatter.string(from: Date())
    
    // Create route, taking care to encode special characters
    let task = task.addingPercentEncoding(withAllowedCharacters: CharacterSet.urlQueryAllowed)!
    let created = date.addingPercentEncoding(withAllowedCharacters: CharacterSet.urlQueryAllowed)!
    let route = "todos?task=\(task)&done=false&created=\(created)"
    
    // Send request, loading items when it's done.
    self.sendRequest("POST", route) {
      self.loadItems()
    }
  }
  
  
  /**
   Toggle the done status of an item
   
   **Modifies**: Self
   
   **Effects**: Flips the done status of the item on the server, updates display.
   
   -Parameter item: The todo item that we are modifying.
   */
  func toggleStatus(item: ToDoItem) {
    let task = item.task.addingPercentEncoding(withAllowedCharacters: CharacterSet.urlQueryAllowed)!
    let created = item.created.addingPercentEncoding(withAllowedCharacters: CharacterSet.urlQueryAllowed)!
    let route = "todos?id=\(item.id)&task=\(task)&done=\(!item.done)&created=\(created)"
    self.sendRequest("PUT", route) {
      self.loadItems()
    }
  }
  

  /**
   Remove an item from our list on the server.
   
   **Modifies**: Self
   
   **Effects**: Removes the item on the server, updates display.
   
   -Parameter item: The todo item to remove.
   */
  func remove(item : ToDoItem) {
    let route = "todos?id=\(item.id)"
    sendRequest("DELETE", route) {
      self.loadItems()
    }
  }
  
  // MARK: - UI Actions
  
  /**
   Prompt user to enter a new task, and create a new todo item for it.
   
   **Modifies**: Self
   
   **Effects**: Adds a new item to the list if user enters a task.
   
   */
  @IBAction func addItem(_ sender: UIBarButtonItem) {
    let controller = UIAlertController(title: "Create New Task", message: nil, preferredStyle: .alert)
    controller.addTextField(configurationHandler: { $0.placeholder = "Task" })
    controller.addAction(UIAlertAction(title: "Done", style: .default) {
      _ in
      
      if let task = controller.textFields![0].text,
         task.count > 0 {
        self.add(task: task)
      }
    })
    self.present(controller, animated: false, completion: nil )
    
  }
  
  // MARK: - REST request sender
  
  /**
   Send a request via http to the server and run the completion
   handler when the respone comes back or we get an error.
   
   **Requires**: http method is GET,POST,PUT,DELETE.
   
   - Parameter httpMethod: One of GET,POST,PUT,DELETE
   - Parameter route: The url of the endpoint we wish to connect to.
   
   - Parameter completion: The code to run when we are done with
   the request.
   */
  private func sendRequest(_ httpMethod: String, _ route : String, _ completion: @escaping () -> Void) {
    
    // construct the http request, including the method
    let url = URL(string: "http://\(host):\(port)/\(route)")!
    var request = URLRequest(url: url)
    request.httpMethod = httpMethod
    
    // Use the shared session to run the request as a data task.
    // Not guaranteed to run the handler on originating thread.
    let task = URLSession.shared.dataTask(with: request) {
      data, response, error in
      if error == nil,
        let httpStatus = response as? HTTPURLResponse,
        httpStatus.statusCode == 200 {
        
        // no error -- good to go!
        completion()
        
      } else {
        
        // error -- see if we can get a real message or
        // make a general one
        let message = error?.localizedDescription ?? "Request Failed"
        print(message)
        
      }
    }
    
    // this is what actually causes the request to run.
    task.resume()
  }
  

  
  // MARK: - Table view data source
  
  /// -Returns: number of sections in the table (always 1).
  override func numberOfSections(in tableView: UITableView) -> Int {
    return 1
  }
  
  /// -Returns: number of todo items in the table.
  override func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
    return items.count
  }
  
  /// -Returns: the table view cell for the item, with text in gray if the item is done.
  override func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
    let cell = tableView.dequeueReusableCell(withIdentifier: "Item", for: indexPath)
    
    let item = items[indexPath.row]
    
    cell.textLabel?.text = item.task
    cell.detailTextLabel?.text = item.created
    
    let color = item.done ? UIColor.lightGray : UIColor.darkText
    cell.textLabel?.textColor = color
    cell.detailTextLabel?.textColor = color
    
    return cell
  }
  
  /**
   Toggles done status of todo item.
   
   **Modifies**: Self
   
   **Effects**: Toggles done status of item at indexPath.row
   
   -Parameter indexPath: The path that was selected.
   */
  override func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
    let item = items[indexPath.row]
    self.toggleStatus(item: item)
  }

  /**
   Removes an item in response to a table delete operation.
   
   **Modifies**: Self
   
   **Effects**: Deletes the item at indexPath.row index.
   
   -Parameter indexPath: The path of the item that should be deleted.
   */
  override func tableView(_ tableView: UITableView, commit editingStyle: UITableViewCellEditingStyle, forRowAt indexPath: IndexPath) {
    if editingStyle == .delete {
      let item = items[indexPath.row]
      self.remove(item: item)
    }
  }

  /// On refresh, just re-load the todo items.
  @IBAction func refresh(_ sender: UIRefreshControl) {
    loadItems()
  }
  
}

