//
//  ToDoDataBase.swift
//  ToDoServer
//
//  Created by Stephen Freund on 11/18/18.
//  Copyright © 2018 Stephen Freund. All rights reserved.
//

import Foundation
import SQLite

/**
 A Database-backed list of ToDoItems.
 */
public class ToDoService {
  
  /// The connection to the database -- create once and use for
  /// all queries.
  private let db : Connection!
  
  // The table and columns we have.
  private let todos = Table("todos")
  private let id = Expression<Int>("id")
  private let task = Expression<String>("task")
  private let created = Expression<String>("created")
  private let done = Expression<Bool>("done")
  
  /// Creates the new ToDoService, and initializes the database table
  /// if it does not already exist.
  public init() {
    do {
      
      // Create DB in user's doc directory.
      let path = NSSearchPathForDirectoriesInDomains(
        .documentDirectory, .userDomainMask, true).first!
      
      self.db = try? Connection("\(path)/db.sqlite3")
      
      // CREATE TABLE "todos" (
      //     "id"      INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
      //     "task"    TEXT NOT NULL,
      //     "created" TEXT NOT NULL,
      //     "done"    BOOLEAN NOT NULL
      // )
      try db.run(todos.create(ifNotExists: true) { t in
        t.column(id, primaryKey: .autoincrement)
        t.column(task)
        t.column(created)
        t.column(done)
      })
    } catch let e {
      assertionFailure("Couldn't create Database connection: \(e)")
    }
  }
  
  /// The items...
  public var items : [ToDoItem] {
    // SELECT * from "todos"
    if let rows = try? db.prepare(todos) {
      // convert each database Row object into a ToDoItem
      return rows.map { (row : Row) in
        ToDoItem(id: row[id],
                 task: row[task],
                 done: row[done],
                 created: row[created])
      }
    } else {
      return []
    }
  }
  
  
  /**
   Create a new task.
   
   **Modifies**: Self
   
   **Effects**: Adds a new task to items
   
   - Parameter task: the task in question.
   - Parameter done: Whether it is done or not.
   - Parameter created: When the task was created
   - Returns: true if successful.
   */
  public func insert(task : String, done : Bool, created : String) -> Bool {
    // INSERT INTO "todos" ("task", "done", "created") VALUES (task, done, created)"
    let insert = todos.insert(self.task <- task,
                              self.done <- done,
                              self.created <- created)
    if let _ = try? db.run(insert) {
      return true
    } else {
      return false
    }
  }
  
  
  /**
   - Returns: the ToDoItem with the given id, or nil if not found.
   */
  public func find(id toDoId: Int) -> ToDoItem? {
    do {
      // SELECT * FROM "users"
      // WHERE ("id" = toDoId)
      // LIMIT 1
      if let row = try db.pluck(todos.filter(id == toDoId)) {
        return ToDoItem(id: row[id],
                        task: row[task],
                        done: row[done],
                        created: row[created])
      } else {
        return nil
      }
    } catch _ {
      return nil
    }
  }
  
  /**
   Updates an existing task.
   
   **Modifies**: Self
   
   **Effects**: Modifies the task with the given id
   
   - Parameter id: unique id of the task we wish to update
   - Parameter task: the description.
   - Parameter done: Whether it is done or not.
   - Parameter created: When the task was created.
   - Returns: true if the task was found and updated, and false otherwise.
   */
  public func update(id toDoId: Int,
                     task : String,
                     done : Bool,
                     created : String) -> Bool {
    do {
      // UPDATE "todos" SET "task", "done", "created" = task, done, created
      // WHERE ("id" = toDoId)
      try db.run(todos.filter(id == toDoId).update(self.task <- task,
                                                   self.done <- done,
                                                   self.created <- created))
      return true
    } catch _ {
      return false
    }
  }
  
  /**
   Removes an existing task.
   
   **Modifies**: Self
   
   **Effects**: Removes the task with the given id
   
   - Parameter id: unique id of the task we wish to remove
   - Returns: true if the task was found and deleted, and false otherwise.
   */
  public func delete(id toDoId: Int) -> Bool {
    do {
      // DELETE FROM "todos" WHERE ("id" = toDiId)
      return try db.run(todos.filter(id == toDoId).delete()) != 0
    } catch _ {
      return false
    }
  }
}
