//
//  TodoServerTableViewController.swift
//  ToDoServer
//
//  Created by Stephen Freund on 11/17/18.
//  Copyright © 2018 Stephen Freund. All rights reserved.

import UIKit
import Swifter
import SQLite


/**
 The server for our ToDo App.  This server recognizes the endpoints specified
 in REST.md.  It uses an SQL database to store todo items.
 
 This server makes use of two additional frameworks:
 
 - SQLite: Swift wrappers for an sqllite database.
 - Swifter: an HTTP Server.  I have modified Swifter in several small ways to
 better support logging of requests/responses.
 
 */
class TodoServerTableViewController: UITableViewController, HttpServerIODelegate {
  
  /// The http server that handles network requests
  private let server: HttpServer = HttpServer()
  
  /// Our "service" that manages a list of todo items
  private let todos : ToDoService = ToDoService()
  
  // MARK: - ViewController Lifecycle
  
  /**
   Sets up the controller after it is loaded.
   
   **Modifies**: Self
   
   **Effects**: Installs handlers for endpoints and starts the server.
   */
  override func viewDidLoad() {
    super.viewDidLoad()
    
    server.delegate = self
    
    server.POST["/todos"]   = { self.add($0) }
    server.GET["/todos"]    = { self.items($0) }
    server.PUT["/todos"]    = { self.update($0) }
    server.DELETE["/todos"] = { self.delete($0) }
    
    do {
      try server.start(9080)
    } catch let e {
      print("Failed to start server: \(e)")
      abort()
    }
  }
  
  // MARK: - Endpoint handlers
  
  /**
   Adds a new todo item.
   
   **Modifies**: Self
   
   **Effects**: Inserts a new todo item into our list, using the
   information encoded in the request parameters
   
   - Parameter request: The request given to us by the server.
   - Returns: The response we create, either .ok, or .badRequest...
   */
  private func add(_ request: HttpRequest) -> HttpResponse {
    let params = request.queryParamsAsMap
    if let task = params["task"],
      let done = Bool(params["done"] ?? ""),
      let created = params["created"],
      todos.insert(task: task, done: done, created: created) {
      return .ok(.json(Data()))
    } else {
      return .badRequest(nil)
    }
  }
  
  /**
   Create a reponse with the list of all to do items.
   
   - Parameter request: The request given to us by the server.
   - Returns: The response we create, with a payload containing
   the JSON for the todo items.
   */
  private func items(_ request: HttpRequest) -> HttpResponse {
    let encoder = JSONEncoder()
    encoder.outputFormatting = JSONEncoder.OutputFormatting.prettyPrinted
    return .ok(.json(try! encoder.encode(todos.items)))
  }
  
  /**
   Updates an existing todo item.
   
   **Modifies**: Self
   
   **Effects**: Removes a todo item in our list
   
   - Parameter request: The request given to us by the server.
   - Returns: Either .ok, or .badRequest if
   item does not exist.
   */
  private func update(_ request: HttpRequest) -> HttpResponse {
    let params = request.queryParamsAsMap
    if let task = params["task"],
      let done = Bool(params["done"] ?? ""),
      let created = params["created"],
      let rowid = Int(params["id"] ?? ""),
      todos.find(id: rowid) != nil,
      todos.update(id: rowid, task: task, done:done, created : created) {
      return .ok(.json(Data()))
    } else {
      return .badRequest(nil)
    }
  }
  
  /**
   Deletes an existing todo item.
   
   **Modifies**: Self
   
   **Effects**: Modifies a todo item in our list, using the
   information encoded in the request parameters
   
   - Parameter request: The request given to us by the server.
   - Returns: The response we create, either .ok, or .badRequest if
   the parameters were ill-formed or the item does not exist.
   */
  private func delete(_ request: HttpRequest) -> HttpResponse {
    let params = request.queryParamsAsMap
    if let rowid = Int(params["id"] ?? ""),
      todos.find(id: rowid) != nil,
      todos.delete(id: rowid) {
      return .ok(.json(Data()))
    } else {
      return .badRequest(nil)
    }
  }
  
  
  
  // MARK: - Logging
  
  /// One request of response.
  private enum LogMessage {
    case request(HttpRequest)
    case response(HttpResponse)
  }
  
  /// An array of all requests/responses processed by the server.
  private var log: [LogMessage] = [] {
    didSet {
      DispatchQueue.main.async {
        self.tableView.reloadData()
        self.tableView.scrollToRow(at: IndexPath(row: self.log.count-1, section: 0), at: .bottom, animated: true)
      }
    }
  }

  // MARK: Delegate methods for HTTP server.

  func socketConnectionReceived(_ socket: Socket) {   }
  
  /// **Effects**: Adds request to log.
  func requestReceived(_ request: HttpRequest) {
    DispatchQueue.main.async {
      self.log.append(.request(request))
    }
  }
  
  /// **Effects**: Adds response to log.
  func responseCreated(_ response: HttpResponse) {
    DispatchQueue.main.async {
      self.log.append(.response(response))
    }
  }
  
  
  
  
  // MARK: - Table view data source
  
  override func numberOfSections(in tableView: UITableView) -> Int {
    return 1
  }
  
  override func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
    return log.count
  }
  
  /// -Returns: The table view cell for the given request/response.  The view
  /// encodes the type of message and ok/bad with colors.
  override func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
    let cell = tableView.dequeueReusableCell(withIdentifier: "Log", for: indexPath)
    
    let row = indexPath.row
    switch(log[row]) {
    case .request(let request):
      cell.textLabel?.text = "\(row) \(request.method): \(request.path)"
      cell.textLabel?.textColor = UIColor.black
    case .response(let response):
      
      switch (response) {
      case .ok(.json(let data)):
        let text = String(data: data, encoding: .utf8) ?? "ERROR"
        cell.textLabel?.text = "\(row) \(response.reasonPhrase()): \(text)"
        cell.textLabel?.textColor = UIColor.blue
      default:
        cell.textLabel?.text = "\(row) \(response.reasonPhrase())"
        cell.textLabel?.textColor = UIColor.red
      }
    }
    
    return cell
    
  }
}





private extension HttpRequest {
  /// The parameters in a request, converting to a map.
  var queryParamsAsMap : [String:String] {
    return Dictionary<String,String>(uniqueKeysWithValues: self.queryParams)
  }
}
